module ccm_barrier_mod
    contains
    subroutine ccm_barrier(wait,io_flush,the_err)
        use ccm_numz
        use ccm_error_mod
        implicit none
        real,optional :: wait
        logical,optional :: io_flush
        integer,optional :: the_err
        integer :: c,r,m,it
        real(b8):: t1,t2,dt,new_t,tx1,tx2
        if(present(the_err))the_err=0
        call MPI_barrier(mycomm,mpi_err);if(mpi_err .ne. 0)call ccm_mpi_error("ccm_barrier")
        if(present(io_flush))then
            if(io_flush)then
!               do i=1,size(ccm_iounits)
!                   call flush(ccm_iounits(i))
!               enddo
                call MPI_barrier(mycomm,mpi_err);if(mpi_err .ne. 0)call ccm_mpi_error("ccm_barrier")
            endif
        endif
        if(present(wait))then
            dt=wait
        else
            return
        endif
        tx1=ccm_time()+dt
        if(dt .gt. 0 .and. dt .lt. 1.0_b8)then
            call system_clock(count=c,count_rate=r,count_max=m)
            t1=real(c,b8)/real(r,b8)
            t2=t1+dt
            new_t=t1
            do while(t2 .gt. new_t)
                call system_clock(count=c)
                new_t=real(c,b8)/real(r,b8)
                if(new_t .lt. t1)then
                    new_t=real(m,b8)-new_t
                    t1=new_t
                    t2=real(m,b8)-t2
                endif
            enddo
        else
            if(dt .gt. 0.0_b8)call sleep(nint(dt))
        endif
        tx2=ccm_time()
        do while(tx2 .lt. tx1)
                tx2=ccm_time()
        enddo
        call MPI_barrier(mycomm,mpi_err);if(mpi_err .ne. 0)call ccm_mpi_error("ccm_barrier")
    end subroutine
    
    subroutine ccm_checkin(wait,name,the_err)
        use ccm_numz
        use ccm_error_mod
        implicit none
        integer :: status(mpi_status_size)
        real,optional,intent(in) :: wait
        character(len=*),intent(in),optional :: name
        integer, optional :: the_err
        real(b8) :: wtime,st,tn,et
        integer :: ierr,igot,i,from,len
        logical :: flag
        character(len=32)  :: lname,gname
        if(present(the_err))the_err=0
        if(present(wait))then
            if(wait .le. 0)return
            wtime=wait
        else
            wtime=10.0
        endif
        if(present(name))then
            if(len_trim(name) .le. 32)then
                write(lname,"(a)")trim(name)
            else
                write(lname,"(a)")name(1:32)
            endif
        else
            write(lname,"(a)")"ccm_undefined"
        endif
        lname=trim(lname)
        len=len_trim(lname)
        st=ccm_time()
        tn=st
        et=st+wtime
        !write(*,"(i4,2f15.3)")myid,st,et
        if(myid .eq. 0)then
            igot=1
            do while(igot .lt. numnodes .and. tn .lt. et)
                flag=.false.
                call  MPI_Iprobe(MPI_ANY_SOURCE, 8765, mycomm, flag, status,ierr )
                if(flag)then
                    from=status(mpi_source)
                     call mpi_recv(gname,32,mycharacter,from,8765,mycomm,status,ierr )
                     if(lge(gname(1:len),lname(1:len)) .and. lle(gname(1:len),lname(1:len)))then
                         igot=igot+1
                         !write(*,*)"igot=",igot
                     else
                        write(err_str1,*)"deadlock detected at ",tn," process ",myid," and ",from
                        write(err_str2,*)"waiting for label ",trim(gname)," but got  ",trim(lname)
                        call ccm_warning("ccm_checkin")
                        if(present(the_err))the_err=-1
                        return
                     endif
                endif
                tn=ccm_time()
            enddo
            if(igot .ne. numnodes)then
               write(err_str1,*)"process ",myid," timed out at ",tn," called with label ",trim(lname)
               call ccm_warning("ccm_checkin")
               if(present(the_err))the_err=-1
               return
            endif
            do i=1,numnodes-1
                call mpi_send(lname,32,mycharacter,i,9876,mycomm,ierr)
            enddo
            return
        endif
        call mpi_send(lname,32,mycharacter,0,8765,mycomm,ierr)
        igot=0
        do while(igot .lt. 1 .and. tn .lt. et)
                flag=.false.
                call  MPI_Iprobe(MPI_ANY_SOURCE, 9876, mycomm, flag, status,ierr )
                if(flag)then
                    from=status(mpi_source)
                     call mpi_recv(gname,32,mycharacter,from,9876,mycomm,status,ierr )
                     if(lge(gname(1:len),lname(1:len)) .and. lle(gname(1:len),lname(1:len)))then
                         igot=igot+1
                     else
                        write(err_str1,*)"deadlock detected at ",tn," process ",myid
                        write(err_str2,*)"waiting for label ",trim(gname)," but got ",trim(lname)
                        call ccm_warning("ccm_checkin")
                        if(present(the_err))the_err=-1
                        return
                     endif
                endif
                tn=ccm_time()
         enddo
         if(igot .ne. 1)then
               write(err_str1,*)"process ",myid," timed out at ",tn," called with label ",trim(lname)
               call ccm_warning("ccm_checkin")
               if(present(the_err))the_err=-1
               return
         endif
       
    end subroutine 
end module
