module ccm_error_mod
    contains
            subroutine ccm_clear_warning()
              use ccm_numz
              write(err_str1,"(1a1)")" "
              write(err_str2,"(1a1)")" "
              write(err_str3,"(1a1)")" "
              write(err_str4,"(1a1)")" "
            end subroutine ccm_clear_warning
            subroutine ccm_warning(routine,dtype,rank1,rank2)
            use ccm_numz
            character(len=*),intent(in):: routine
            integer,optional :: dtype
            integer,intent(in),optional::rank1,rank2
            character(len=25):: the_type
                write(the_type,*)"undefined"
                if(present(dtype))then
                    write(the_type,*)"nonstandard"
                    if(dtype .eq. myreal)write(the_type,*)"real single precision"
                    if(dtype .eq. mydouble)write(the_type,*)"real double precision"
                    if(dtype .eq. mycomp)write(the_type,*)"complex single precision"
                    if(dtype .eq. mydpcomp)write(the_type,*)"complex double precision"
                    if(dtype .eq. myint)write(the_type,*)"normal integer"
                    if(dtype .eq. mycharacter)write(the_type,*)"character"
                    if(dtype .eq. mylogical)write(the_type,*)"logical"
                    if(dtype .eq. myquad)write(the_type,*)"real quad precision"
                    if(dtype .eq. myqpcomp)write(the_type,*)"complex quad precision"
                    if(dtype .eq. mylong)write(the_type,*)"alternate length integer"
                endif
                
            if(present(dtype))then
                write(err_str0,*)"routine: ",routine," and data type ",the_type
            else
                write(err_str0,*)"routine: ",routine
            endif
            if(present(rank1))then
                if(rank1 .gt. -1)then
                    write(err_str3,*)"rank of input data:",rank1
                else
                    write(err_str3,*)"rank of input data: undefined"
                endif
            else
                write(err_str3,*)" "
            endif
            
            if(present(rank2))then
                if(rank2 .gt. -1)then
                    write(err_str4,*)"rank of output data:",rank2
                else
                    write(err_str4,*)"rank of output data: undefined"
                endif
            else
                write(err_str4,*)" "
            endif
            if(ccm_auto_print)then
                write(*,"(i5,a,a)")myid,": ","Warning from collective communications module"
                write(*,"(i5,a,a)")myid,": ",trim(err_str0)
                write(*,"(i5,a,a)")myid,": ",trim(err_str1)
                write(*,"(i5,a,a)")myid,": ",trim(err_str2)
                write(*,"(i5,a,a)")myid,": ",trim(err_str3)
                write(*,"(i5,a,a)")myid,": ",trim(err_str4)
            endif
        end subroutine
        subroutine ccm_print_warning(file_number)
            use ccm_numz
            integer, optional :: file_number
            if(present(file_number))then
                write(file_number,*)"Warning from collective communications module"
                write(file_number,"(i5,a,a)")myid,": ",trim(err_str0)
                write(file_number,"(i5,a,a)")myid,": ",trim(err_str1)
                write(file_number,"(i5,a,a)")myid,": ",trim(err_str2)
                write(file_number,"(i5,a,a)")myid,": ",trim(err_str3)
                write(file_number,"(i5,a,a)")myid,": ",trim(err_str4)
            else
                write(*,*)"Warning from collective communications module"
                write(*,"(i5,a,a)")myid,": ",trim(err_str0)
                write(*,"(i5,a,a)")myid,": ",trim(err_str1)
                write(*,"(i5,a,a)")myid,": ",trim(err_str2)
                write(*,"(i5,a,a)")myid,": ",trim(err_str3)
                write(*,"(i5,a,a)")myid,": ",trim(err_str4)
            endif
        end subroutine  ccm_print_warning       
        subroutine ccm_fatal(routine,dtype,rank1,rank2)
        use ccm_numz
            character(len=*),intent(in):: routine
            integer,optional :: dtype
            integer,intent(in),optional::rank1,rank2
            character(len=25):: the_type
                write(the_type,*)"undefined"
                if(present(dtype))then
                    if(dtype .eq. myreal)write(the_type,*)"real single precision"
                    if(dtype .eq. mydouble)write(the_type,*)"real double precision"
                    if(dtype .eq. mycomp)write(the_type,*)"complex single precision"
                    if(dtype .eq. mydpcomp)write(the_type,*)"complex double precision"
                    if(dtype .eq. myint)write(the_type,*)"normal integer"
                    if(dtype .eq. mycharacter)write(the_type,*)"character"
                    if(dtype .eq. mylogical)write(the_type,*)"logical"
                    if(dtype .eq. mylong)write(the_type,*)"alternate length integer"
                endif
                
            write(*,*)"fatal error from collective communications module"
            if(present(dtype))then
                write(err_str0,*)"routine: ",routine," and data type ",the_type
            else
                write(err_str0,*)"routine: ",routine
            endif
            write(*,"(i5,a,a)")myid,": ",err_str0
            if(present(rank1))then
                if(rank1 .gt. -1)then
                    write(*,"(i5,a,a)")myid,": ","rank of input data:",rank1
                else
                    write(*,"(i5,a,a)")myid,": ","rank of input data: undefined"
                endif
            endif
            if(present(rank2))then
                if(rank2 .gt. -1)then
                    write(*,"(i5,a,a)")myid,": ","rank of output data:",rank2
                else
                    write(*,"(i5,a,a)")myid,": ","rank of output data: undefined"
                endif
            endif
            write(*,"(i5,a,a)")myid,": ",trim(err_str1)
            write(*,"(i5,a,a)")myid,": ",trim(err_str2)
            call mpi_abort(mpi_comm_world,-1,mpi_err)
            stop
        end subroutine
        subroutine ccm_mpi_error(routine,details)
            use ccm_numz
            character(len=*),intent(in):: routine
            character(len=*),optional,intent(in):: details
    		write(err_str1,*)"underling communications error"
    		if(present(details))write(err_str2,*)details
    		call ccm_fatal(routine)
    	end subroutine ccm_mpi_error

end module
